// ==============================================================
// WARP MFD - tested only in the Solar System!
// ==============================================================

#define STRICT
#define ORBITER_MODULE
#include "stdio.h"
#include "stdlib.h"
#include "random.h"
#include "windows.h"
#include "orbitersdk.h"
#include "WarpMFD.h"

// ==============================================================
// Global variables

int g_MFDmode; // identifier for new MFD mode
MFDWarp *warpInstance; //Instance of MFD

//Some strings
const char* strings[5]= {
	"Warp MFD",
	"Distance to travel:\0",
	"Final distance to target:\0",
	"Final absolute velocity:\0",
	"Final relative velocity:\0"
};


// ==============================================================
// API interface

DLLCLBK void opcDLLInit (HINSTANCE hDLL)
{
	static char *name = "Warp";       // MFD mode name
	MFDMODESPEC spec;
	spec.name = name;
	spec.key = OAPI_KEY_W;            // MFD mode selection key
	spec.msgproc = MFDWarp::MsgProc;  // MFD mode callback function

	warpInstance = NULL;

	// Register the new MFD mode with Orbiter
	g_MFDmode = oapiRegisterMFDMode (spec);
}

DLLCLBK void opcDLLExit (HINSTANCE hDLL)
{
	// Unregister the custom MFD mode when the module is unloaded
	oapiUnregisterMFDMode (g_MFDmode);
}

DLLCLBK void opcTimestep (double SimT, double SimDT, double mjd) {
	static int t=-1;
	// Dummy cleanup after warp jump
	if (warpInstance && warpInstance->warpCleanup) {
		if (t==-1) t = 2;

		if (t--==0) {
			// Unlock the focus vessel from the dummy object
			warpInstance->dummyInt->DetachChild(warpInstance->attachDummy);
	
			// Delete the dummy object
			oapiDeleteVessel(warpInstance->dummy);

			warpInstance->warpCleanup = false;
		}
	}
}

// ==============================================================
// MFD class constructor and destructor

// Constructor
MFDWarp::MFDWarp (DWORD w, DWORD h, VESSEL *vessel)
: MFD (w, h, vessel)
{
//	if (!warpInstance)
		warpInstance = this; // set the current instance

	selectTarget("", false); // No target
	userDistance = 1000000000L; // Default: warp 10^6 km away!
	userVelocity = 0;

	dummy = NULL;
	warpCleanup = false;
	success = true;
	calcsDone = false;

	sprintf(debugString,"debug: inited");
}

// Destructor
MFDWarp::~MFDWarp()
{
	warpInstance = NULL;
}

// ==============================================================
// USER interface

// Protypes of Callbacks for Input Boxes
bool cbSelectTarget(void *id, char *str, void *data);
bool cbDistance(void *id, char *str, void *data);
bool cbVelocity(void *id, char *str, void *data);
bool cbConfirmWarp(void *id, char *str, void *data);

// Return button labels
char *MFDWarp::ButtonLabel (int bt)
{
	// The labels for the buttons used by the MFD 
	static char *label[5] = {"TGT", "NT", "DIS", "VEL", "WRP"};
	return (bt < 5 ? label[bt] : 0);
}

// Return button menus
int MFDWarp::ButtonMenu (const MFDBUTTONMENU **menu) const
{
	// The menu descriptions for the buttons
	static const MFDBUTTONMENU mnu[5] = {		
		{"<- Select target, final distance",
		 "  to target, and final velocity.", 'T'},
		{"<- No target. Trajectory in the",
		 "  ship direction.", 'N'},
		{"<- Distance to travel w/o target",
		 "  or final distance to target.", 'D'},
		{"<- Select final velocity in km/s",
		 " (will be in ship direction)", 'V'},
		{"<- Do the hyperspace jump",
		 0, 'W'},
	};
	if (menu) *menu = mnu;
	return 5; // return the number of buttons used
}

bool MFDWarp::ConsumeButton (int bt, int event) {

	if (!(event & PANEL_MOUSE_LBDOWN)) return false;

	switch (bt) {
		case 0: ConsumeKeyBuffered(OAPI_KEY_T); break; // Target
		case 1: selectTarget("", true); break;		   // No target
		case 2: ConsumeKeyBuffered(OAPI_KEY_D); break; // Distance
		case 3: ConsumeKeyBuffered(OAPI_KEY_V); break; // Vel
		case 4: ConsumeKeyBuffered(OAPI_KEY_W); break; // Warp
		default: return false;
	}
	return true;
}

bool MFDWarp::ConsumeKeyBuffered(DWORD key)
{
	switch(key) {
	case OAPI_KEY_T:
		oapiOpenInputBox("(Warp): Select Target (leave blank for none)", cbSelectTarget, 0, 20, (void *)this);
		return true;
	case OAPI_KEY_N:
		selectTarget("", true);
		return true;
	case OAPI_KEY_D:
		if (target==NULL)
			oapiOpenInputBox("(Warp): Select distance to travel (preceed with K, M, G or A )", cbDistance, 0, 20, (void *)this);
		else
			oapiOpenInputBox("(Warp): Select final distance from target (preceed with % or K, M, G, A)", cbDistance, 0, 20, (void *)this);
		return true;
	case OAPI_KEY_V:
		oapiOpenInputBox("(Warp): Select final velocity (Km/s)", cbVelocity, 0, 20, (void *)this);
		return true;
	case OAPI_KEY_W:
		// Recalculate "warp trajectory" and errors
		if (!target)
			selectTarget("", true);
		else
			selectTarget(targetName, true);
		success = true;
		char s[100];
		sprintf(s,"Probability of safe warp: %.1f%% Proceed? (y/n)", (1-probFailure)*100);
		oapiOpenInputBox(s, cbConfirmWarp, 0, 3, (void *)this);
		return true;
	default:
		return false;
	}

}

#define LINE 16
void print(HDC hDC, int i, int *j, LPCTSTR cad) {
	TextOut(hDC, i, *j+=LINE, cad, sizeof(char)*strlen(cad));
}

void sprintDistance(char *s, double dist, bool AU) {
	if (dist >= 1000000000L) {
		dist /= 1000000000L;
		if (AU)
			sprintf(s, "%.2f Gm (%.3f AU)", dist, dist / 150);
		else
			sprintf(s, "%.2f Gm", dist);
		return;
	}
	if (dist >= 1000000L) {
		dist /= 1000000L;
		if (AU)
			sprintf(s, "%.2f Mm (%.3f AU)", dist, dist / 150000L);
		else
			sprintf(s, "%.2f Mm", dist);
		return;
	}	
	if (dist >= 1000L) {
		dist /= 1000L;
		if (AU)
			sprintf(s, "%.2f Km (%.3f AU)", dist, dist / 150000000L);
		else
			sprintf(s, "%.2f Km", dist);
		return;
	}	
	sprintf(s, "%.1f m", dist);
}

// ==============================================================
// Repaint the MFD
void MFDWarp::Update (HDC hDC)
{

	Title (hDC, strings[0]);

	int line = 32;
	char s[100];	
	#define RESET_LINE line = 32;

	// Update target info
	if (target)	selectTarget(targetName, false);


	// Print target
	sprintf(s,"[Shift-T] Target: %s", targetName);
	print(hDC, 10, &line, s);
	if (target) {
		sprintf(s,"Curr. distance: ");
		sprintDistance(s+LINE, targetDistance, true);
		print(hDC, 20, &line, s);
		sprintf(s,"Curr. velocity: %.3f Km/s",targetVelocity/1000);
		print(hDC, 20, &line, s);
		// Print radius
		sprintf(s,"Radius: ");
		sprintDistance(s+8, targetRadius, false);
		print(hDC, 20, &line, s);
	} else line+=LINE*3;
	line+=LINE;

	// Print distance
	sprintf(s,"[Shift-D] %s", (target?strings[2]:strings[1]));
	print(hDC, 10, &line, s);
	sprintDistance(s, userDistance, true);
	print(hDC, 20, &line, s);
	line+=LINE;

	// Print velocity
	sprintf(s,"[Shift-V] %s", (target?strings[4]:strings[3]));
	print(hDC, 10, &line, s);
	sprintf(s,"%.3f Km/s", userVelocity/1000);
	print(hDC, 20, &line, s);
	line+=LINE;

	// Print warp success probability if calculated
	if (calcsDone) {
		sprintf(s,"Probability of safe warp: %.1f%%", (1-probFailure)*100);
		print(hDC, 40, &line, s);
	}
	line+=LINE;

	// Print warp result if failure
	if (!success) {
		sprintf(s,"<<<Bad jump!>>>");
		print(hDC, 40, &line, s);
	}

	// Print debug
//	print(hDC, 10, &line, debugString);
}

// ==============================================================
// MFD message parser
int MFDWarp::MsgProc (UINT msg, UINT mfd, WPARAM wparam, LPARAM lparam)
{
	switch (msg) {
	case OAPI_MSG_MFD_OPENED:
		// Our new MFD mode has been selected, so we create the MFD and
		// return a pointer to it.
		return (int)(new MFDWarp (LOWORD(wparam), HIWORD(wparam), (VESSEL*)lparam));
	}
	return 0;
}


// ==============================================================
// Internal functionality

// selectTarget(char *name)
//
// Select a target by name and do the "warp computing" of the trajectory, wich
// has an error in distance and final velocity. 
//
// The mass bodies near the straight line from the vessel to the target affects
// the hiperspace navigation in a complex way. 
// In case of bad jump, the errors grow exponentially. The computer throws a rough estimation
// of the probability that a gravitational influence has not been taked into account. 
// Interestellar and even interplanetary travels make easy to not have been detected a small
// body, and this especially occur near the destination.
//
bool MFDWarp::selectTarget(char *name, bool doCalcs) {
	if (strcmp(name,"")==0) {
		strcpy(targetName, "None");
		target = NULL;
	}
	else {
		OBJHANDLE t = oapiGetObjectByName(name);
		if (t != NULL) {
			oapiGetObjectName(t,targetName,NAME_SIZE);
			target = t;
			oapiGetFocusRelativePos(target, &targetRelPos);
			oapiGetFocusRelativeVel(target, &targetRelVel);
			targetDistance = length(targetRelPos);
			targetVelocity = length(targetRelVel);
			targetRadius = oapiGetSize(t);
		}
		else return false;
	}

	// Calculate "warp trajectory"
	if (doCalcs) {
		if (!calculateWarpPos()) return true;
		// (todo): Put hiper-spatial theory stuff here 
		#define NP 10
		VECTOR3 pos[NP];
		double g[NP], gm = 0;

		VECTOR3 wd = warpPos - focusPos;
		DWORD i, n;
		for (i = 0; i < NP; i++) {
			pos[i] = focusPos + (wd*( ((double)i) / ((double)(NP-1)) ));
			g[i] = 0;
		}
		n = oapiGetGbodyCount();
		for (i = 0; i < n; i++) {
			OBJHANDLE hBody = oapiGetGbodyByIndex (i);
			VECTOR3 r; oapiGetGlobalPos(hBody, &r);
			double mass = oapiGetMass(hBody);
			for (int j = 0; j < NP; j++) {
				VECTOR3 a = pos[j]-r;
				double d2 = a.x*a.x + a.y*a.y + a.z*a.z;
				if (d2 > 10) {
					g[j] +=  0.0000000000667 * mass / d2;
					if (i==(n-1) && g[j] > gm) gm = g[j];
				}
			}
		}

		probFailure = gm<2? gm/2 : 1;
		if ((targetDistance - userDistance) < (targetRadius + oapiGetSize(focus)))
			probFailure = 0;
		calcsDone = true;
	}

	return true;
}

// Calculate warp jump exit position
bool MFDWarp::calculateWarpPos(void) {
	focus = oapiGetFocusInterface();
	if (!focus) return false;
	focus->GetGlobalPos(focusPos);
	focus->GlobalRot(_V(0,0,1),focusDir);
	if (target==NULL) {
		VECTOR3 v = focusDir * userDistance;
		warpDist = length(v);
		warpPos = focusPos + v;
		warpVel = focusDir * userVelocity;
	} else {
		if (userDistance >= targetDistance) return false;

		double d = (targetDistance - userDistance) / targetDistance;
		VECTOR3 v = targetRelPos * d;
		warpDist = length(v);
		warpPos = focusPos - v;
		oapiGetGlobalVel(target, &warpVel);
		warpVel += focusDir * userVelocity;
	}
	return true;
}

// Perform the warp jump
void MFDWarp::doWarp(void) {	

	if (!calculateWarpPos()) return;

	// Hiperspace error
	randomize();
	double e = 0.5;frand;
	success = (e >= probFailure);
	if (!success) {
		e = e-probFailure;
		if (e==0) e = 0.000001;
		e = warpDist / e;
		warpPos+= _V(noise(e),noise(e),noise(e));
	}

	// Create dummy vessel (with a 'tochild' attachment) in the warp jump exit position
	VESSELSTATUS2 vs;
	vs.version = 2;
	vs.flag = 0;
	vs.rbody = oapiGetGbodyByIndex(0);
	vs.base = NULL;
	vs.port = 0;
	vs.status = 0;
	veccpy(vs.rpos, warpPos);
	veccpy(vs.rvel, warpVel);
	vs.vrot = _V(0,0,0);
	vs.arot = _V(0,0,0);
	vs.nfuel = 0;
	vs.fuel = NULL;
	vs.nthruster = 0;
	vs.thruster = NULL;
	vs.ndockinfo = 0;
	vs.dockinfo = NULL;
	vs.xpdr = 0;

	dummy = oapiCreateVesselEx("warp_dummy", "warpdummy", &vs);
	dummyInt = oapiGetVesselInterface(dummy);
	attachDummy = dummyInt->CreateAttachment(false, _V(0,0,0),_V(0,0,-1),_V(0,1,0), "DUMMY", true);

	// If there is no "DUMMY" child attachment in the focus vessel, create one
	ATTACHMENTHANDLE attachFocus;
	int i = 0, n = focus->AttachmentCount(true);
	while (i<n) {
		attachFocus = focus->GetAttachmentHandle(true, i);
		if (strcmp(focus->GetAttachmentId(attachFocus),"DUMMY")==0) break;
		i++;
	}
	if (i==n) {
		attachFocus = focus->CreateAttachment(true, _V(0,0,0), _V(0,0,1), _V(0,1,0), "DUMMY", true);
	}

	// Link the focus vessel to the dummy, performing the warp jump
	dummyInt->AttachChild(oapiGetFocusObject(), attachDummy, attachFocus);

	// (we are right now in hyperspace...)
	calcsDone = false;

	// Do the cleanup in the next frame 
	warpCleanup = true;
}

// ==============================================================
// Input Boxes Callback functions
bool cbSelectTarget(void *id, char *str, void *data)
{
	MFDWarp *w = (MFDWarp *)data;
	w->calcsDone= false;
	return w->selectTarget(str, true);
}

bool cbDistance(void *id, char *str, void *data) {
	MFDWarp *w = (MFDWarp *)data;
	if (strlen(str)==0) return false;	

	double d, mult = 1;
	switch(*str) {
		case '%':
			if (!w->target) return false; 
			str++;
			if (strlen(str)==0) return false;
			d = atof(str);
			if (d<0 || d>100) return false;
			w->userDistance = w->targetDistance*(d/100.0);			
			mult = 0;
			break;
		case 'k':
		case 'K':
			mult = 1000L;
			break;
		case 'm':
		case 'M':
			mult = 1000000L;
			break;
		case 'g':
		case 'G':
			mult = 1000000000L;
			break;
		case 'a':
		case 'A':
			mult = 149597870691L;
			break;
	}
	if (mult > 1) {
		str++;
		if (strlen(str)==0) return false;
	}
	if (mult != 0) {
		d = atof(str) * mult;
		if (d<0) return false;
		w->userDistance = d;
	}
	w->calcsDone= false;
	// Recalculate "warp trajectory" and errors
	if (!w->target)
		w->selectTarget("", true);
	else
		w->selectTarget(w->targetName, true);
	return true;
}

bool cbVelocity(void *id, char *str, void *data) {
	MFDWarp *w = (MFDWarp *)data;
	w->userVelocity = atof(str) * 1000;
	w->calcsDone= false;
	// Recalculate "warp trajectory" and errors
	if (!w->target)
		w->selectTarget("", true);
	else
		w->selectTarget(w->targetName, true);
	return true;
}

bool cbConfirmWarp(void *id, char *str, void *data) {
	MFDWarp *w = (MFDWarp *)data;
	if (strcmp(str,"y")==0 || strcmp(str,"Y")==0) {
		w->doWarp();
	}
	return true;
}